/***************************************************************************
* Copyright (c) 2015-2018 Pier Luigi Fiorini <pierluigi.fiorini@gmail.com>
* Copyright (c) 2013 Abdurrahman AVCI <abdurrahmanavci@gmail.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the
* Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
***************************************************************************/

#include "GreeterApp.h"
#include "GreeterProxy.h"
#include "Constants.h"
#include "ScreenModel.h"
#include "UserModel.h"
#include "KeyboardModel.h"
#include "MessageHandler.h"

#include <QCommandLineParser>
#include <QGuiApplication>
#include <QQuickItem>
#include <QQmlContext>
#include <QQmlEngine>
#include <QDebug>
#include <QTimer>
#include <QTranslator>
#include <sys/types.h>
#include <signal.h>
#include <statgrab.h>
#include "Utils.h"
#include <QX11Info>
#include <sys/prctl.h>
#include <QDBusInterface>
#include <QEvent>
#include <KDeclarative/KQuickAddons/QuickViewSharedEngine>
static const QEvent::Type StartupEventType = static_cast<QEvent::Type>(QEvent::registerEventType());

class FocusOutEventFilter : public QAbstractNativeEventFilter
{
public:
    bool nativeEventFilter(const QByteArray &eventType, void *message, long int *result) override {
        Q_UNUSED(result)
        if (qstrcmp(eventType, "xcb_generic_event_t") != 0) {
            return false;
        }
        xcb_generic_event_t *event = reinterpret_cast<xcb_generic_event_t*>(message);
        if ((event->response_type & ~0x80) == XCB_FOCUS_OUT) {
            return true;
        }
        return false;
    }
};
namespace NSS {
GreeterApp::GreeterApp(int &argc, char **argv)
    : QGuiApplication(argc, argv)
{
    m_components_tranlator = new QTranslator();
    if (m_components_tranlator->load(QLocale::system(), QString(), QString(), QStringLiteral(COMPONENTS_TRANSLATION_DIR)))
        QCoreApplication::installTranslator(m_components_tranlator);

    m_userModel = new UserModel();
    m_appModel = new AppsModel();
    m_appsIcon = new AppsIconProvider();
    m_keyboard = new KeyboardModel();
    if (QX11Info::isPlatformX11()) {
        installNativeEventFilter(new FocusOutEventFilter);
    }
}
GreeterApp::~GreeterApp() {
    for(KQuickAddons::QuickViewSharedEngine *view : m_views) {
        view->setKeyboardGrabEnabled(false);
        view->setMouseGrabEnabled(false);
        if (QQuickItem *focusItem = view->activeFocusItem()) {
            focusItem->setFocus(false);
        }
    }
    qDeleteAll(m_views);
    m_userModel->deleteLater();
    m_appModel->deleteLater();
    m_appModel->deleteLater();
    m_keyboard->deleteLater();
}

QString GreeterApp::socketName() const
{
    return m_socket;
}

void GreeterApp::setSocketName(const QString &name)
{
    m_socket = name;
}
void GreeterApp::customEvent(QEvent *event)
{
    if (event->type() == StartupEventType)
        startup();
}
bool GreeterApp::eventFilter(QObject *obj, QEvent *event)
{
    Q_UNUSED(obj)
    if (event->type() == QEvent::MouseButtonPress && QX11Info::isPlatformX11()) {
        KQuickAddons::QuickViewSharedEngine *activeScreen = getActiveScreen();
        if(activeScreen) {
            activeScreen->requestActivate();
        }
    }
    if ( event->type() == QEvent::KeyPress ||
         event->type() == QEvent::KeyRelease ) {
        if (static_cast<QKeyEvent *>(event)->key() == Qt::Key_Escape) {
            return true;
        }
    }
    return false;
}
KQuickAddons::QuickViewSharedEngine *GreeterApp::getActiveScreen()
{
    KQuickAddons::QuickViewSharedEngine *activeScreen = nullptr;

    if (m_views.isEmpty()) {
        return activeScreen;
    }

    for (KQuickAddons::QuickViewSharedEngine *view : m_views) {
        if (view->geometry().contains(QCursor::pos())) {
            activeScreen = view;
            break;
        }
    }
    if (!activeScreen) {
        activeScreen = m_views.first();
    }

    return activeScreen;
}
void GreeterApp::addViewForScreen(QScreen *screen) {
    // create view
    auto *view = new KQuickAddons::QuickViewSharedEngine();
    view->setScreen(screen);
    view->setFlags(Qt::X11BypassWindowManagerHint);
    view->setResizeMode(KQuickAddons::QuickViewSharedEngine::SizeRootObjectToView);
    view->setGeometry(screen->geometry());

    m_views.append(view);

    connect(qGuiApp, &QGuiApplication::screenRemoved, view, [view, this, screen](QScreen *s) {
        if (s == screen)
            removeViewForScreen(view);
    });

    // always resize when the screen geometry changes
    connect(screen, &QScreen::geometryChanged, this, [view](const QRect &r) {
        view->setGeometry(r);
    });
    if(!m_imageProvider) {
        view->engine()->addImportPath(QStringLiteral(IMPORTS_INSTALL_DIR));
        view->engine()->addImageProvider(QStringLiteral("appsIcon"),m_appsIcon);
        m_imageProvider = true;
    }


    // connect proxy signals
    connect(m_proxy, &GreeterProxy::loginSucceeded, view, &KQuickAddons::QuickViewSharedEngine::close);
    view->rootContext()->setContextProperty(QStringLiteral("userModel"), m_userModel);
    view->rootContext()->setContextProperty(QStringLiteral("appModel"), m_appModel);
    view->rootContext()->setContextProperty(QStringLiteral("nss"), m_proxy);
    view->rootContext()->setContextProperty(QStringLiteral("keyboard"), m_keyboard);
    view->rootContext()->setContextProperty(QStringLiteral("primaryScreen"), QGuiApplication::primaryScreen() == screen);
    view->rootContext()->setContextProperty(QStringLiteral("__nss_errors"), QString());
    view->rootContext()->setContextProperty(QStringLiteral("winId"), static_cast<quint64>(view->winId()));

    // get theme main script
    QUrl mainScriptUrl = QUrl::fromLocalFile(":/theme/Main.qml");
    // set main script as source
    view->setSource(mainScriptUrl);

    // set default cursor
    QCursor cursor(Qt::ArrowCursor);
    view->rootObject()->setCursor(cursor);

    view->show();
    view->raise();
    view->setKeyboardGrabEnabled(true);
    view->requestActivate();
}
void GreeterApp::removeViewForScreen(KQuickAddons::QuickViewSharedEngine *view) {
    // screen is gone, remove the window
    m_views.removeOne(view);
    view->deleteLater();
}

void GreeterApp::startup()
{
    m_proxy = new GreeterProxy(m_userModel->currentUser(), m_socket);
    if (!m_proxy->isConnected()) {
        qCritical() << "Cannot connect to the daemon - is it running?";
        QCoreApplication::exit(EXIT_FAILURE);
        return;
    }
    m_keyboard->setNumLockState(true);
    installEventFilter(this);
    QList<QScreen *> screens = qGuiApp->primaryScreen()->virtualSiblings();
    for (QScreen *screen : screens) {
//        qWarning() << screen;
        addViewForScreen(screen);
    }
    connect(qGuiApp, &QGuiApplication::screenAdded, this, &GreeterApp::addViewForScreen);
    connect(qGuiApp, &QGuiApplication::primaryScreenChanged, this, [this](QScreen *) {
        activatePrimary();
    });
}
void GreeterApp::activatePrimary() {
    Q_FOREACH (KQuickAddons::QuickViewSharedEngine *view, m_views) {
        if (view->screen() == QGuiApplication::primaryScreen()) {
            view->requestActivate();
            break;
        }
    }
}
StartupEvent::StartupEvent()
    : QEvent(StartupEventType)
{
}
}
static void signalHandler(int signum)
{
    NSS::GreeterApp *instance = qobject_cast<NSS::GreeterApp *>(QCoreApplication::instance());

    if (!instance)
        return;

    switch(signum)
    {
      case SIGTERM:
        instance->exit(1);
        break;
    }
}
static void startCompositer()
{
    QDBusInterface * compositer = new QDBusInterface(QString::fromLatin1("org.nde.session"),
                                                     QString::fromLatin1("/NdeSession"),
                                                     QString::fromLatin1("org.nde.session"),
                                                     QDBusConnection::sessionBus());
    if(compositer->isValid())
    {
        QDBusMessage reply = compositer->call("getCompositorOpen");
        if(!reply.arguments().isEmpty() && !reply.arguments().first().toBool())
        {
            qputenv("QMLSCENE_DEVICE", "softwarecontext");
        }
    }
    delete compositer;
}
int main(int argc, char **argv)
{
    qputenv("QT_IM_MODULE", QByteArray("qtvirtualkeyboard"));
    startCompositer();
    qInstallMessageHandler(NSS::GreeterMessageHandler);
    QGuiApplication::setAttribute(Qt::AA_EnableHighDpiScaling);

    NSS::GreeterApp app(argc, argv);
    QCommandLineParser parser;
    parser.setApplicationDescription(TR("nde-screensaver greeter"));
    parser.addHelpOption();
    parser.addVersionOption();

    QCommandLineOption socketOption(QLatin1String("socket"), TR("Socket name"), TR("name"));
    parser.addOption(socketOption);
    parser.process(app);

    app.setSocketName(parser.value(socketOption));
    QCoreApplication::postEvent(&app, new NSS::StartupEvent());
    struct sigaction sa;
    sa.sa_handler = signalHandler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;
    sigaction(SIGTERM, &sa, nullptr);
    return app.exec();
}
